# -*- coding: utf-8 -*-
"""
Created on Wed Jul  5 13:12:34 2017

@author: eyt21

This module contains the application to stream and record data.
It contains widgets to configure and toggle channel plots,
configure recording device, and configure recording.
It displays a time domain plot, a frequecy domain plot (DFT),
and a channel levels plot, all in real time.

Further information on the functionality of each widget can be found in
RecordingUI.

The application is meant to be used with the analysis window application,
because recorded data are transfer to that window for analysis.

However, the application can work standalone, but the recorded data will
go nowhere.

Attribute
----------
NI_DRIVERS: bool
    Indicates whether NIDAQmx drivers and pyDAQmx module are installed
    when attempting to import NIRecorder module
    These are required to interface with a National Instrument hardware.
PLAYBACK: bool
    Indicates whether to play the streaming audio, only works with SoundCard
WIDTH: Int
    Width of the application window
HEIGHT: Int
    Height of the application window
"""
import sys,traceback
from PyQt5.QtWidgets import (QWidget,QHBoxLayout,QMainWindow,QPushButton,
                             QDesktopWidget,QRadioButton,QSplitter,
                             QApplication)
from PyQt5.QtCore import Qt, QTimer, pyqtSignal

import pyqtgraph as pg

import copy
import numpy as np
from numpy.fft import rfft

from cued_datalogger.acquisition.RecordingUIs import (ChanToggleUI,ChanConfigUI,DevConfigUI,
                                                 StatusUI,RecUI)
from cued_datalogger.acquisition.RecordingGraph import TimeLiveGraph,FreqLiveGraph,LevelsLiveGraph
from cued_datalogger.acquisition.ChanMetaWin import ChanMetaWin

import cued_datalogger.acquisition.myRecorder as mR
try:
    import cued_datalogger.acquisition.NIRecorder as NIR
    NI_drivers = True
except NotImplementedError:
    print("Seems like you don't have National Instruments drivers")
    NI_drivers = False
except ImportError:
    print("ImportError: Seems like you don't have pyDAQmx modules")
    NI_drivers = False
from cued_datalogger.analysis.frequency_domain import (compute_transfer_function,
                                                   compute_autospec,compute_crossspec)

from cued_datalogger.api.channel import ChannelSet
from cued_datalogger.api.toolbox import Toolbox, MasterToolbox

# GLOBAL CONSTANTS
PLAYBACK = False    # Whether to playback the stream
WIDTH = 900         # Window width
HEIGHT = 600        # Window height

#++++++++++++++++++++++++ The AcquisitionWindow Class +++++++++++++++++++++++++++
class AcquisitionWindow(QMainWindow):
    """
    This is the window that acts as the central hub for all the widgets.
    Contains a left Toolbox for streaming configuration and a right Toolbox
    for recording configuration.
    The center contain the Time Domain and Frequency Domain Plots, along with a
    status bar below them. The channel levels plot is placed in the right Toolbox.

    The widgets communicate with each other using the pyqt Signal and Slot method,
    so most callback functions are self-contained in the respective widget.
    However, some callback function are still present in the main body, mostly
    functions to handle the streaming and the recording processes.

    Attributes
    ----------
    dataSaved: pyqtSignal
        Emitted when recorded data
    done: pyqtSignal
        Emitted when the window is closed
    playing: bool
        Indicate whether the stream is playing
    rec: Recorder object
         Object which handles the streaming and recording
         See documentation for Recorder classes
    """
    dataSaved = pyqtSignal(int)
    done = pyqtSignal()

    def __init__(self,parent = None,recType = mR,configs = ['',44100,2,1024,6]):
        """
        Initialise and construct the window application.

        Parameters
        ----------
        parent: object
            The object which the window interacts with
            e.g. The analysis window
        """
        super().__init__()
        self.parent = parent

        # Set window parameter
        self.setGeometry(400,300,WIDTH,HEIGHT)
        self.setWindowTitle('AcquisitionWindow')

        self.meta_window = None

        # Set recorder object
        self.playing = False
        self.rec = recType.Recorder(rate = configs[1],
                                    channels = configs[2],
                                    chunk_size = configs[3],
                                    num_chunk = configs[4],
                                    device_name = configs[0])
        # Set up the TimeSeries and FreqSeries
        self.timedata = None
        self.freqdata = None

        # Set up tallies for the average transfer function calculation
        self.autospec_in_tally = []
        self.autospec_out_tally = []
        self.crossspec_tally = []

        try:
            # Construct UI
            self.initUI()
        except Exception:
            # If it fails, show the window and stop
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            self.show()
            return

        # Connect the recorder Signals
        self.connect_rec_signals()

        # Attempt to start streaming
        self.init_and_check_stream()

        # Center and show window
        self.adjust_position()
        self.setFocus()
        self.show()

    def initUI(self):
        """
        Construct the window by calling the widgets classes, and any additional
        widgets (such as Qsplitters, Toolbox)
        Then, perform the signal connection among the widgets.
        Also sets up the update timer for the streaming.
        Lastly, finalise the plots and misc. items.

        There is also an experimental styling section just after the UI
        construction.
        """
        # Set up the main widget
        self.main_widget = QWidget(self)
        main_layout = QHBoxLayout(self.main_widget)
    #------------------------- STREAM TOOLBOX ------------------------------
        self.stream_toolbox = MasterToolbox()
        self.stream_tools = Toolbox('left',self.main_widget)
        self.stream_toolbox.add_toolbox(self.stream_tools)
        self.chantoggle_UI = ChanToggleUI(self.main_widget)
        self.chanconfig_UI = ChanConfigUI(self.main_widget)
        self.devconfig_UI = DevConfigUI(self.main_widget)
        self.devconfig_UI.set_recorder(self.rec)
        self.devconfig_UI.config_setup()
        NI_btn = self.devconfig_UI.typegroup.findChildren(QRadioButton)[1]
        if not NI_drivers:
            NI_btn.setDisabled(True)
        self.stream_tools.addTab(self.chantoggle_UI,'Channel Toggle')
        self.stream_tools.addTab(self.chanconfig_UI,'Channel Config')
        self.stream_tools.addTab(self.devconfig_UI,'Device Config')
        main_layout.addWidget(self.stream_toolbox)
        main_layout.setStretchFactor(self.stream_toolbox, 0)

    #---------------------------PLOT + STATUS WIDGETS-----------------------------
        self.mid_splitter = QSplitter(self.main_widget,orientation = Qt.Vertical)
        self.timeplot = TimeLiveGraph(self.mid_splitter)
        self.freqplot = FreqLiveGraph(self.mid_splitter)
        self.stats_UI = StatusUI(self.mid_splitter)
        self.mid_splitter.addWidget(self.timeplot)
        self.mid_splitter.addWidget(self.freqplot)
        self.mid_splitter.addWidget(self.stats_UI)
        self.mid_splitter.setCollapsible (2, False)
        main_layout.addWidget(self.mid_splitter)
        main_layout.setStretchFactor(self.mid_splitter, 1)

    #---------------------------RECORDING TOOLBOX-------------------------------
        self.recording_toolbox = MasterToolbox()
        self.recording_tools = Toolbox('right',self.main_widget)
        self.recording_toolbox.add_toolbox(self.recording_tools)
        self.right_splitter = QSplitter(self.main_widget,orientation = Qt.Vertical)
        self.RecUI = RecUI(self.main_widget)
        self.RecUI.set_recorder(self.rec)
        self.right_splitter.addWidget(self.RecUI)
        self.levelsplot = LevelsLiveGraph(self.rec,self.right_splitter)
        self.right_splitter.addWidget(self.levelsplot)
        self.recording_tools.addTab(self.right_splitter,'Record Time Series')
        main_layout.addWidget(self.recording_toolbox)
        main_layout.setStretchFactor(self.recording_toolbox, 0)

    #-----------------------EXPERIMENTAL STYLING----------------------------
        #self.main_splitter.setFrameShape(QFrame.Panel)
        #self.main_splitter.setFrameShadow(QFrame.Sunken)
        self.main_widget.setStyleSheet('''
        .QWidget{
            background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                stop:0 #eee, stop:1 #ccc);
            border: 1px solid #777;
            width: 13px;
            margin-top: 2px;
            margin-bottom: 2px;
            border-radius: 4px;
        }
        .QSplitter::handle{
                background: #737373;
        }
        .QGroupBox{
                border: 1px solid black;
                margin-top: 0.5em;
                font: italic;
        }
        .QGroupBox::title {
                top: -6px;
                left: 10px;
        }
        .QWidget #subWidget{
                background: qlineargradient(x1:0, y1:0, x2:1, y2:1,
                    stop:0 #eee, stop:1 #ccc);
                border: 1px solid #777;
                width: 13px;
                margin-top: 2px;
                margin-bottom: 2px;
                border-radius: 4px;
            }
        ''')

    #----------------------SIGNAL CONNECTIONS---------------------------
        self.chantoggle_UI.sigToggleChanged.connect(self.timeplot.toggle_plotline)
        self.chantoggle_UI.sigToggleChanged.connect(self.freqplot.toggle_plotline)
        self.chanconfig_UI.chans_num_box.currentIndexChanged.connect(self.display_chan_config)
        self.chanconfig_UI.meta_btn.clicked.connect(self.open_meta_window)
        self.chanconfig_UI.sigTimeOffsetChanged.connect(self.timeplot.set_offset)
        self.chanconfig_UI.sigFreqOffsetChanged.connect(self.freqplot.set_offset)
        self.chanconfig_UI.sigHoldChanged.connect(self.timeplot.set_sig_hold)
        self.chanconfig_UI.sigColourChanged.connect(self.timeplot.set_plot_colour)
        self.chanconfig_UI.sigColourChanged.connect(self.freqplot.set_plot_colour)
        self.chanconfig_UI.sigColourChanged.connect(self.levelsplot.set_plot_colour)
        self.chanconfig_UI.sigColourReset.connect(self.timeplot.reset_default_colour)
        self.chanconfig_UI.sigColourReset.connect(self.freqplot.reset_default_colour)
        self.chanconfig_UI.sigColourReset.connect(self.levelsplot.reset_default_colour)
        self.devconfig_UI.configRecorder.connect(self.ResetRecording)
        self.timeplot.plotColourChanged.connect(self.chanconfig_UI.set_colour_btn)
        self.freqplot.plotColourChanged.connect(self.chanconfig_UI.set_colour_btn)
        self.levelsplot.plotColourChanged.connect(self.chanconfig_UI.set_colour_btn)
        self.levelsplot.thresholdChanged.connect(self.RecUI.rec_boxes[4].setText)
        self.stats_UI.statusbar.messageChanged.connect(self.default_status)
        self.stats_UI.resetView.pressed.connect(self.ResetSplitterSizes)
        self.stats_UI.togglebtn.pressed.connect(lambda: self.toggle_rec())
        self.stats_UI.sshotbtn.pressed.connect(self.get_snapshot)
        self.RecUI.rec_boxes[4].textEdited.connect(self.levelsplot.change_threshold)
        self.RecUI.startRecording.connect(self.start_recording)
        self.RecUI.cancelRecording.connect(self.cancel_recording)
        self.RecUI.undoLastTfAvg.connect(self.undo_tf_tally)
        self.RecUI.clearTfAvg.connect(self.remove_tf_tally)
    #---------------------------RESETTING METHODS---------------------------
        self.ResetMetaData()
        self.ResetChanBtns()
        self.ResetPlots()
        self.ResetChanConfigs()
        self.levelsplot.reset_channel_levels()
        self.ResetSplitterSizes()
    #-----------------------FINALISE THE MAIN WIDGET-------------------------
        #Set the main widget as central widget
        self.main_widget.setFocus()
        self.setCentralWidget(self.main_widget)

        # Set up a timer to update the plot
        self.plottimer = QTimer(self)
        self.plottimer.timeout.connect(self.update_line)
        #self.plottimer.timeout.connect(self.update_chanlvls)
        self.plottimer.start(self.rec.chunk_size*1000//self.rec.rate)

        self.show()

    def adjust_position(self):
        """
        If it has a parent, adjust its position to be slightly out of the parent
        window towards the left
        """
        if self.parent:
            pr = self.parent.frameGeometry()
            qr = self.frameGeometry()
            self.move(pr.topLeft())
            self.move(qr.left()/2,qr.top())

#----------------CHANNEL CONFIGURATION WIDGET---------------------------
    def display_chan_config(self, arg):
        """
        Displays the channel plot offsets, colours, and signal holding.
        """
        # This function is here because it is easier for channel config to
        # get data from the plot widgets
        if type(arg) == pg.PlotDataItem:
            num = self.timeplot.check_line(arg)
            if not num == None:
                num = self.freqplot.check_line(arg)
                self.chanconfig_UI.chans_num_box.setCurrentIndex(num)
        else:
            num = arg

        self.chanconfig_UI.colbox.setColor(self.timeplot.plot_colours[num])
        self.chanconfig_UI.time_offset_config[0].setValue(self.timeplot.plot_xoffset[num])
        self.chanconfig_UI.time_offset_config[1].setValue(self.timeplot.plot_yoffset[num])
        self.chanconfig_UI.hold_tickbox.setCheckState(self.timeplot.sig_hold[num])
        self.chanconfig_UI.fft_offset_config[0].setValue(self.freqplot.plot_xoffset[num])
        self.chanconfig_UI.fft_offset_config[1].setValue(self.freqplot.plot_yoffset[num])

    def open_meta_window(self):
        """
        Open the metadata window
        """
        if not self.meta_window:
            try:
                self.meta_window = ChanMetaWin(self)
                self.meta_window.show()
                self.meta_window.finished.connect(self.meta_win_closed)
            except:
                t,v,tb = sys.exc_info()
                print(t)
                print(v)
                print(traceback.format_tb(tb))

    def meta_win_closed(self):
        """
        Callback when the metadata window is closed
        """
        self.meta_window = None
        self.update_chan_names()

#----------------------PLOT WIDGETS-----------------------------------
    # Updates the plots
    def update_line(self):
        """
        Callback to update the time domain and frequency domain plot
        """
        # Get the buffer
        data = self.rec.get_buffer()

        # Take the last chunk for the levels plot
        currentdata = data[len(data)-self.rec.chunk_size:,:]
        currentdata -= np.mean(currentdata)
        rms = np.sqrt(np.mean(currentdata ** 2,axis = 0))
        maxs = np.amax(abs(currentdata),axis = 0)
        self.levelsplot.set_channel_levels(rms,maxs)

        # Prepare the window and weightage for FFT plot
        window = np.hanning(data.shape[0])
        weightage = np.exp(2* self.timedata / self.timedata[-1])

        # Update each plot item's data + level peaks
        for i in range(data.shape[1]):
            plotdata = data[:,i].reshape((len(data[:,i]),))

            # Check for zero crossing if needed
            zc = 0
            if self.timeplot.sig_hold[i] == Qt.Checked:
                avg = np.mean(plotdata);
                zero_crossings = np.where(np.diff(np.sign(plotdata-avg))>0)[0]
                if zero_crossings.shape[0]:
                    zc = zero_crossings[0]+1

            self.timeplot.update_line(i,x = self.timedata[:len(plotdata)-zc] ,y = plotdata[zc:])
            fft_data = rfft(plotdata* window * weightage)
            psd_data = abs(fft_data)** 0.5
            self.freqplot.update_line(i,x = self.freqdata ,y = psd_data)
            self.levelsplot.set_peaks(i,maxs[i])

#-------------------------STATUS BAR WIDGET--------------------------------
    def toggle_rec(self,stop = None):
        """
        Callback to pause/resume the stream, unless explicitly specified to stop or not

        Parameters
        ----------
        stop: bool
            Specify whether to stop the stream. None to toggle the current state
        """
        if not stop == None:
            self.playing = stop

        if self.playing:
            self.rec.stream_stop()
            self.stats_UI.togglebtn.setText('Resume')
            self.RecUI.recordbtn.setDisabled(True)
        else:
            self.rec.stream_start()
            self.stats_UI.togglebtn.setText('Pause')
            self.RecUI.recordbtn.setEnabled(True)
        self.playing = not self.playing
        # Clear the status, allow it to auto update itself
        self.stats_UI.statusbar.clearMessage()

    # Get the current instantaneous plot and transfer to main window
    def get_snapshot(self):
        """
        Callback to take the current buffer data and send it out to parent window
        """
        snapshot = self.rec.get_buffer()
        for i in range(snapshot.shape[1]):
            self.live_chanset.set_channel_data(i,'time_series',snapshot[:,i])

        self.live_chanset.set_channel_metadata( tuple(range(snapshot.shape[1])),
                                                   {'sample_rate':self.rec.rate})
        self.save_data()
        self.stats_UI.statusbar.showMessage('Snapshot Captured!', 1500)

    def default_status(self,msg):
        """
        Callback to set the status message to the default messages
        if it is empty (ie when cleared)
        """
        # Placed here because it accesses self.playing (might change it soon?)
        if not msg:
            if self.playing:
                self.stats_UI.statusbar.showMessage('Streaming')
            else:
                self.stats_UI.statusbar.showMessage('Stream Paused')


#---------------------------RECORDING WIDGET-------------------------------
    def start_recording(self):
        """
        Callback to start the data recording
        """
        success = False
        rec_configs = self.RecUI.get_record_config()
        if rec_configs[2]>=0:
            # Set up the trigger if specified
            if self.rec.trigger_start(posttrig = rec_configs[0],
                                      duration = rec_configs[1],
                                      pretrig = rec_configs[2],
                                      channel = rec_configs[3],
                                      threshold = rec_configs[4]):
                success = True
                self.stats_UI.statusbar.showMessage('Trigger Set!')
        else:
            # Record normally
            self.rec.record_init(samples = rec_configs[0], duration = rec_configs[1])
            if self.rec.record_start():
                success = True
                self.stats_UI.statusbar.showMessage('Recording...')
        if success:
            # Disable buttons
            for btn in [self.stats_UI.togglebtn, self.devconfig_UI.config_button, self.RecUI.recordbtn]:
                btn.setDisabled(True)
            self.RecUI.switch_rec_box.setDisabled(True)
            self.RecUI.spec_settings_widget.setDisabled(True)
            # Enable the cancel buttons
            self.RecUI.cancelbtn.setEnabled(True)

    #
    def stop_recording(self):
        """
        Callback when the recording finishes.
        Process the data first, if specify, then transfer the data
        to the parent window
        """
        # Enable the buttons
        for btn in self.main_widget.findChildren(QPushButton):
            btn.setEnabled(True)
        # Disable the cancel button
        self.RecUI.cancelbtn.setDisabled(True)

        # Get the recorded data and compute DFT
        data = self.rec.flush_record_data()
        ft_datas = np.zeros((int(data.shape[0]/2)+1,data.shape[1]),dtype = np.complex)
        for i in range(data.shape[1]):
            self.live_chanset.set_channel_data(i,'time_series',data[:,i])
            ft = rfft(data[:,i])
            self.live_chanset.add_channel_dataset(i,'spectrum',ft)
            ft_datas[:,i] = ft

        self.live_chanset.set_channel_metadata( tuple(range(data.shape[1])),
                                                   {'sample_rate':self.rec.rate})

        # Check recording mode
        rec_mode = self.RecUI.get_recording_mode()
        if rec_mode == 'Normal':
            # Send data normally
            self.live_chanset.add_channel_dataset(tuple(range(data.shape[1])),'TF',[])
            self.live_chanset.add_channel_dataset(tuple(range(data.shape[1])),'coherence',[])
            self.save_data(0)
        elif rec_mode == 'TF Avg.':
            # Compute the auto- and crossspectrum for average transfer function
            # while store them in the tally
            chans = list(range(self.rec.channels))
            in_chan = self.RecUI.get_input_channel()
            chans.remove(in_chan)
            input_chan_data = ft_datas[:,in_chan]

            # Check for incorrect data length with previous recorded data
            if not len(self.autospec_in_tally)==0:
                if not input_chan_data.shape[0] == self.autospec_in_tally[-1].shape[0]:
                    print('Data shape does not match, you may have fiddle the settings')
                    print('Please either clear the past data, or revert the settings')
                    self.stats_UI.statusbar.clearMessage()
                    self.RecUI.spec_settings_widget.setEnabled(True)
                    self.RecUI.switch_rec_box.setEnabled(True)
                    return

            self.autospec_in_tally.append(compute_autospec(input_chan_data))

            autospec_out = np.zeros((ft_datas.shape[0],ft_datas.shape[1] - 1),dtype = np.complex)
            crossspec = np.zeros(autospec_out.shape,dtype = np.complex)
            for i,chan in enumerate(chans):
                autospec_out[:,i] = compute_autospec(ft_datas[:,chan])
                crossspec[:,i] = compute_crossspec(input_chan_data,ft_datas[:,chan])

            self.autospec_out_tally.append(autospec_out)
            self.crossspec_tally.append(crossspec)
            auto_in_sum = np.array(self.autospec_in_tally).sum(axis = 0)
            auto_out_sum = np.array(self.autospec_out_tally).sum(axis = 0)
            cross_sum = np.array(self.crossspec_tally).sum(axis = 0)
            for i,chan in enumerate(chans):
                tf_avg,cor = compute_transfer_function(auto_in_sum,auto_out_sum[:,i],cross_sum[:,i])
                self.live_chanset.add_channel_dataset(chan,'TF',tf_avg)
                self.live_chanset.add_channel_dataset(chan,'coherence',cor)

            # Update the average count and send the data
            self.RecUI.update_TFavg_count(len(self.autospec_in_tally))
            self.save_data(1)

        elif rec_mode == 'TF Grid':
            # TODO: Implement recording for grid transfer function
            pass
        else:
            # TODO?: Failsafe for unknown mode
            pass

        # Enable more widgets
        self.stats_UI.statusbar.clearMessage()
        self.RecUI.spec_settings_widget.setEnabled(True)
        self.RecUI.switch_rec_box.setEnabled(True)

    def undo_tf_tally(self):
        """
        Callback to remove the last autospectrum and crossspectrum in the tally
        """
        if self.autospec_in_tally:
            self.autospec_in_tally.pop()
            self.autospec_out_tally.pop()
            self.crossspec_tally.pop()
        self.RecUI.update_TFavg_count(len(self.autospec_in_tally))

    def remove_tf_tally(self):
        """
        Callback to clear the autospectrum and crossspectrum tallies
        """
        if self.autospec_in_tally:
            self.autospec_in_tally = []
            self.autospec_out_tally = []
            self.crossspec_tally = []
        self.RecUI.update_TFavg_count(len(self.autospec_in_tally))

    # Cancel the data recording
    def cancel_recording(self):
        """
        Callback to cancel the recording and re-enable the UIs
        """
        self.rec.record_cancel()
        for btn in self.main_widget.findChildren(QPushButton):
            btn.setEnabled(True)

        self.RecUI.switch_rec_box.setEnabled(True)
        self.RecUI.spec_settings_widget.setEnabled(True)
        self.RecUI.cancelbtn.setDisabled(True)
        self.stats_UI.statusbar.clearMessage()

#--------------------------- RESET METHODS-------------------------------------
    def ResetRecording(self):
        """
        Reset the stream to the specified configuration, then
        calls upon other reset methods
        """
        # Spew errors to console at each major step of the resetting
        self.stats_UI.statusbar.showMessage('Resetting...')

        # Stop the update and Delete the stream
        self.playing = False
        self.plottimer.stop()
        self.rec.close()
        del self.rec

        try:
            # Get Input from the Device Configuration UI
            Rtype, settings = self.devconfig_UI.read_device_config()
            # Reinitialise the recording object
            self.rec = Rtype.Recorder()
            # Set the recorder parameters
            dev_name = self.rec.available_devices()[0]
            sel_ind = min(settings[0],len(dev_name)-1)
            self.rec.set_device_by_name(dev_name[sel_ind])
            self.rec.rate = settings[1]
            self.rec.channels = settings[2]
            self.rec.chunk_size = settings[3]
            self.rec.num_chunk = settings[4]
            self.devconfig_UI.configboxes[0].setCurrentIndex(dev_name.index(self.rec.device_name))
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            print('Cannot set up new recorder')

        try:
            # Open the stream
            self.init_and_check_stream()
            # Reset channel configs
            self.ResetPlots()
            self.levelsplot.reset_channel_peaks(self.rec)
            self.ResetChanConfigs()
            self.levelsplot.reset_channel_levels()
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            print('Cannot stream,restart the app')

        try:
            # Set the recorder reference to RecUI and devconfig_UI, and remove
            # average transfer function tally
            self.RecUI.set_recorder(self.rec)
            self.devconfig_UI.set_recorder(self.rec)
            self.remove_tf_tally()
        except:
            t,v,tb = sys.exc_info()
            print(t)
            print(v)
            print(traceback.format_tb(tb))
            print('Cannot recording configs')

        # Reset metadata and change channel toggles and re-connect the signals
        self.ResetMetaData()
        self.ResetChanBtns()
        self.connect_rec_signals()

        # Restart the plot update timer
        self.plottimer.start(self.rec.chunk_size*1000//self.rec.rate)

    def ResetPlots(self):
        """
        Reset the plots
        """
        self.ResetXdata()

        self.timeplot.reset_plotlines()
        self.freqplot.reset_plotlines()

        for i in range(self.rec.channels):
            tplot = self.timeplot.plot()
            tplot.sigClicked.connect(self.display_chan_config)

            fplot = self.freqplot.plot()
            fplot.sigClicked.connect(self.display_chan_config)

        self.freqplot.plotItem.setRange(xRange = (0,self.freqdata[-1]),yRange = (0, 100*self.rec.channels))
        self.freqplot.plotItem.setLimits(xMin = 0,xMax = self.freqdata[-1],yMin = -20)

    def ResetXdata(self):
        """
        Reset the time and frequencies plot data
        """
        data = self.rec.get_buffer()
        self.timedata = np.arange(data.shape[0]) /self.rec.rate
        self.freqdata = np.arange(int(data.shape[0]/2)+1) /data.shape[0] * self.rec.rate

    def ResetChanBtns(self):
        """
        Reset the amount of channel toggling buttons
        """
        self.chantoggle_UI.adjust_channel_checkboxes(self.rec.channels)
        self.update_chan_names()

    def ResetChanConfigs(self):
        """
        Reset the channel plot configuration settings
        """
        self.timeplot.reset_offsets()
        self.timeplot.reset_plot_visible()
        self.timeplot.reset_colour()
        self.timeplot.reset_sig_hold()
        self.freqplot.reset_offsets()
        self.freqplot.reset_plot_visible()
        self.freqplot.reset_colour()
        self.levelsplot.reset_colour()

        self.chanconfig_UI.chans_num_box.clear()
        self.chanconfig_UI.chans_num_box.addItems([str(i) for i in range(self.rec.channels)])
        self.chanconfig_UI.chans_num_box.setCurrentIndex(0)

        self.display_chan_config(0)

    def ResetMetaData(self):
        """
        Reset the metadata
        """
        self.live_chanset = ChannelSet(self.rec.channels)
        self.live_chanset.add_channel_dataset(tuple(range(self.rec.channels)), 'time_series')

    def ResetSplitterSizes(self):
        """
        Reset the metadata
        """
        self.mid_splitter.setSizes([HEIGHT*0.48,HEIGHT*0.48,HEIGHT*0.04])
        self.right_splitter.setSizes([HEIGHT*0.05,HEIGHT*0.85])

    def update_chan_names(self):
        """
        Update the channel names, obtained from the ChannelSet
        """
        names = self.live_chanset.channel_metadata(tuple(range(self.rec.channels)),'name')
        for n,name in enumerate(names):
            chan_btn = self.chantoggle_UI.chan_btn_group.button(n)
            chan_btn.setText(name)

#----------------------- DATA TRANSFER METHODS -------------------------------

    def save_data(self, tab_num = 0):
        """
        Transfer data to parent window

        Parameters
        ----------
        tab_num: Tab index to switch to in the parent window (specific for analysis window)
        """
        if self.parent:
            print('Saving data...')
            self.parent.cs = copy.copy(self.live_chanset)
            self.dataSaved.emit(tab_num)
            print('Data saved!')

#-------------------------- STREAM METHODS ------------------------------------
    def init_and_check_stream(self):
        """
        Attempts to initialise the stream
        """
        if self.rec.stream_init(playback = PLAYBACK):
            self.stats_UI.togglebtn.setEnabled(True)
            self.toggle_rec(stop = False)
            self.stats_UI.statusbar.showMessage('Streaming')
        else:
            self.stats_UI.togglebtn.setDisabled(True)
            self.toggle_rec(stop = True)
            self.stats_UI.statusbar.showMessage('Stream not initialised!')

    def connect_rec_signals(self):
        """
        Connects the signals from the recorder object
        """
        self.rec.rEmitter.recorddone.connect(self.stop_recording)
        self.rec.rEmitter.triggered.connect(self.stats_UI.trigger_message)
        #self.rec.rEmitter.newdata.connect(self.update_line)
        #self.rec.rEmitter.newdata.connect(self.update_chanlvls)

#----------------------OVERRIDDEN METHODS------------------------------------
    def closeEvent(self,event):
        """
        Reimplemented from QMainWindow
        Stops the update timer, disconnect any signals and delete self
        """
        if self.plottimer.isActive():
            self.plottimer.stop()

        self.done.emit()
        self.rec.close()
        self.dataSaved.disconnect()
        event.accept()
        self.deleteLater()

if __name__ == '__main__':
    app = 0
    app = QApplication(sys.argv)
    w = AcquisitionWindow()
    sys.exit(app.exec_())



import numpy as np

from scipy.signal import ricker, cwt

from cued_datalogger.api.pyqt_widgets import MatplotlibCanvas

from PyQt5.QtWidgets import QApplication, QWidget, QVBoxLayout, QSlider, QLabel, QSpinBox, QHBoxLayout, QGridLayout, QComboBox

from PyQt5.QtCore import Qt

import sys

#--------------------------------
from scipy.signal import gausspulse
#import matplotlib.pyplot as plt

t = np.linspace(-1, 1, 200, endpoint=False)
sig  = np.cos(2 * np.pi * 7 * t) + gausspulse(t - 0.4, fc=2)
widths = np.arange(1, 31)
"""
cwt_result = cwt(sig, ricker, widths)

T, W = np.meshgrid(t, widths)

plt.figure()
plt.contour(T, W, cwt_result)
plt.show()
#----------------------------------
"""


class CWTPlotWidget(MatplotlibCanvas):
    """A MatplotlibCanvas widget displaying the CWT plot"""
    
    def __init__(self, sig, t, widths, wavelet=ricker, plot_type="Colourmap",
                 num_contours=5, contour_spacing_dB=5):
        self.sig = sig
        self.t = t
        self.widths = widths
        self.wavelet = wavelet
        self.plot_type = plot_type
        self.num_contours = num_contours
        self.contour_spacing_dB = contour_spacing_dB
        
        MatplotlibCanvas.__init__(self, "Continuous Wavelet Transform")
        
        self.calculate_cwt()
        self.draw_plot()
    
    def draw_plot(self):
        """Redraw the CWT on the canvas"""
        
        self.axes.clear()
        
        if self.plot_type == "Contour":
            
            self.update_contour_sequence()
            
            self.axes.contour(self.T, self.W, self.cwt_result, self.contour_sequence)
            
            self.axes.set_xlabel('Time (s)')
            self.axes.set_ylabel('Wavelet width (?)')
            
            self.axes.set_xlim(self.t.min(), self.t.max())
            self.axes.set_ylim(self.widths.min(), self.widths.max())
            
        elif self.plot_type == "Surface":
            pass
        
        elif self.plot_type == "Colourmap":
            
            self.update_contour_sequence()
            
            self.axes.pcolormesh(self.T, self.W, self.cwt_result, vmin=self.contour_sequence[0])
            
            self.axes.set_xlabel('Time (s)')
            self.axes.set_ylabel('Wavelet width (?)')
            
            self.axes.set_xlim(self.t.min(), self.t.max())
            self.axes.set_ylim(self.widths.min(), self.widths.max())
            
        else:
            pass
        
        self.draw()
        
    def update_attributes(self, value):
        """A slot for updating the attributes when input widgets are adjusted"""
        
        # What sent the signal?
        sender_name = self.sender().objectName()
        
        if sender_name == "plot_type_combobox":
            self.plot_type = value
        
        elif sender_name == "contour_spacing_spinbox" or sender_name == "contour_spacing_slider":
            self.contour_spacing_dB = value
            
        elif sender_name == "num_contours_spinbox" or sender_name == "num_contours_slider":
            self.num_contours = value
        
        else:
            print("Sender {} not implemented.".format(sender_name))
            # Discard any other signal
            pass
        
        # Update the plot
        self.calculate_cwt()
        self.draw_plot()
    
    def calculate_cwt(self):
        """Recalculate the CWT"""
        self.cwt_result = cwt(self.sig, self.wavelet, self.widths)

        self.T, self.W = np.meshgrid(self.t, self.widths)

    
    def update_contour_sequence(self):
        """Update the array which says where to plot contours, how many etc"""
        # Create a vector with the right spacing from min to max value
        self.contour_sequence = np.arange(self.cwt_result.min(), self.cwt_result.max(),
                                          self.contour_spacing_dB)
        # Take the appropriate number of contours
        self.contour_sequence = self.contour_sequence[-self.num_contours:]


class CWTWidget(QWidget):
    def __init__(self, sig, t, widths=np.arange(1, 31), parent=None):
        self.sig = sig
        self.t = t
        self.widths = widths
               
        super().__init__()
        self.init_ui()
    
    def init_ui(self):
        # Add a cwt plot
        self.cwt_plot = CWTPlotWidget(self.sig, self.t, self.widths)
              
        #------------Plot type controls------------
        self.plot_type_label = QLabel(self)
        self.plot_type_label.setText("Plot type")
        # Create combobox
        self.plot_type_combobox = QComboBox(self)
        self.plot_type_combobox.addItems(["Colourmap", "Contour", "Surface"])
        self.plot_type_combobox.setObjectName("plot_type_combobox")
        # Update on change
        self.plot_type_combobox.activated[str].connect(self.cwt_plot.update_attributes)        
        
        #------------Contour spacing controls------------
        self.contour_spacing_label = QLabel(self)
        self.contour_spacing_label.setText("Contour spacing")
        # Create spinbox       
        self.contour_spacing_spinbox = QSpinBox(self)
        self.contour_spacing_spinbox.setObjectName("contour_spacing_spinbox")        
        self.contour_spacing_spinbox.setRange(1, 12)
        # Create slider        
        self.contour_spacing_slider = QSlider(Qt.Vertical, self)
        self.contour_spacing_slider.setObjectName("contour_spacing_slider")
        self.contour_spacing_slider.setRange(1, 12)
        # Connect spinbox and slider together
        self.contour_spacing_spinbox.valueChanged.connect(self.contour_spacing_slider.setValue)
        self.contour_spacing_slider.valueChanged.connect(self.contour_spacing_spinbox.setValue)
        # Set values
        self.contour_spacing_spinbox.setValue(self.cwt_plot.contour_spacing_dB)
        self.contour_spacing_slider.setValue(self.cwt_plot.contour_spacing_dB)
        # Update screen on change
        self.contour_spacing_slider.valueChanged.connect(self.cwt_plot.update_attributes)
        self.contour_spacing_spinbox.valueChanged.connect(self.cwt_plot.update_attributes)
        
        #------------Num contours controls------------
        self.num_contours_label = QLabel(self)
        self.num_contours_label.setText("Num contours")
        # Create spinbox       
        self.num_contours_spinbox = QSpinBox(self)
        self.num_contours_spinbox.setObjectName("num_contours_spinbox")        
        self.num_contours_spinbox.setRange(1, 12)
        # Create slider        
        self.num_contours_slider = QSlider(Qt.Vertical, self)
        self.num_contours_slider.setObjectName("num_contours_slider")
        self.num_contours_slider.setRange(1, 12)
        # Connect spinbox and slider together
        self.num_contours_spinbox.valueChanged.connect(self.num_contours_slider.setValue)
        self.num_contours_slider.valueChanged.connect(self.num_contours_spinbox.setValue)
        # Set values
        self.num_contours_spinbox.setValue(self.cwt_plot.num_contours)
        self.num_contours_slider.setValue(self.cwt_plot.num_contours)
        # Update screen on change
        self.num_contours_slider.valueChanged.connect(self.cwt_plot.update_attributes)
        self.num_contours_spinbox.valueChanged.connect(self.cwt_plot.update_attributes)
        
        #------------Layout------------
        # CWT controls:
        self.cwt_controls_label = QLabel(self)
        self.cwt_controls_label.setText("<b>CWT controls</b>")
        
        cwt_controls = QGridLayout()
        cwt_controls.addWidget(self.cwt_controls_label, 0, 0)
        
        # Plot controls:
        self.plot_controls_label = QLabel(self)
        self.plot_controls_label.setText("<b>Plot controls</b>")
        
        plot_controls = QGridLayout()
        plot_controls.addWidget(self.plot_controls_label, 0, 0)
        plot_controls.addWidget(self.contour_spacing_label, 1, 0)
        plot_controls.addWidget(self.contour_spacing_spinbox, 2, 0)
        plot_controls.addWidget(self.contour_spacing_slider, 3, 0)
        plot_controls.addWidget(self.num_contours_label, 1, 1)
        plot_controls.addWidget(self.num_contours_spinbox, 2, 1)
        plot_controls.addWidget(self.num_contours_slider, 3, 1)
        plot_controls.addWidget(self.plot_type_label, 4, 0)
        plot_controls.addWidget(self.plot_type_combobox, 4, 1)
                
        vbox = QVBoxLayout()
        hbox = QHBoxLayout()
        
        hbox.addWidget(self.cwt_plot)
        hbox.addLayout(plot_controls)
        
        vbox.addLayout(hbox)
        vbox.addLayout(cwt_controls)

        self.setLayout(vbox)
        self.setWindowTitle('CWT')
        self.show()

if __name__ == '__main__':  
    app = 0
    
    app = QApplication(sys.argv)
    cwt_w = CWTWidget(sig, t)
        
    sys.exit(app.exec_())  
    


# -*- coding: utf-8 -*-
"""
Created on Fri Sep  1 14:00:14 2017

@author: eyt21
"""

import scipy.io as sio
from cued_datalogger.api.channel import ChannelSet
import numpy as np
from PyQt5.QtWidgets import (QWidget, QVBoxLayout,QPushButton,QLabel,QListWidget,
                             QTreeWidgetItem,QHBoxLayout,QFileDialog,QCheckBox)
from PyQt5.QtCore import  Qt
from cued_datalogger.api.numpy_extensions import to_dB
import pickle

def export_to_mat(file,order, channel_set=None,back_comp = False):
    """
    Export data and metadata from ChannelsSet to a mat file.
    Support backward compatibility with the old logger.
    Still in alpha stage.
    
    Parameters
    ----------
    file : path_to_file
        The path to the ``.mat`` file to import data from.
    order : int
        Order of the channel saved
    channel_set : ChannelSet
        The ChannelSet to save the imported data and metadata to. If ``None``, 
        a new ChannelSet is created and returned.
    back_comp: bool
        Whether to export as the old format
    """
    
    if channel_set is None:
        new_channel_set = True
        channel_set = ChannelSet(len(order))
    else:
        new_channel_set = False
    
    # Obtain the ids and metadata names
    data_ids = channel_set.channel_ids(order)
    var_names = set([y for x in data_ids for y in x])
    meta_names = channel_set.channel_metadata(0)
        
    if not back_comp:
        # Save all the variable names and metadata names as dicts
        variables = {}
        for name in var_names:
            data = channel_set.channel_data(order,name)
            variables[name] = data
        for mname in meta_names:
            mdata = channel_set.channel_metadata(order,mname)
            variables[mname] = mdata    
        # Save the dict as matlab file 
        sio.savemat(file,variables,appendmat = False)
        
        if new_channel_set:
            return channel_set
    else:
        # Save as the old format, with individual MAT file for each type of data
        sampling_rate = channel_set.channel_metadata(0,'sample_rate')
        calibration_factor = channel_set.channel_metadata(0,'calibration_factor')
        # Save Time Series
        if 'time_series' in var_names:
            time_series_data = np.array(channel_set.channel_data(order,'time_series'))
            n_samples = time_series_data[0].shape[0]
            variables = {'indata':np.transpose(time_series_data),'freq':float(sampling_rate),
                         'dt2' :[float(len(order)),0,0],'buflen':float(n_samples),
                         'tsmax':float(calibration_factor)}
            time_series_fname = file[:-4]+'_TimeSeries.mat'
            sio.savemat(time_series_fname,variables,appendmat = False)
        # Save FFT
        if 'spectrum' in var_names:
            fft_data = np.array(channel_set.channel_data(order,'spectrum'))
            n_samples = fft_data[0].shape[0]
            variables = {'yspec':np.transpose(fft_data),'freq':float(sampling_rate),
                         'dt2' :[0,float(len(order)),0],'npts':float((n_samples-1)*2),
                         'tfun':0}
            time_series_fname = file[:-4]+'_FFT.mat'
            sio.savemat(time_series_fname,variables,appendmat = False)
        # Save TF
        if 'TF' in var_names:
            TF_data = np.array(channel_set.channel_data(order,'TF'))
            TF_data = [data for data in TF_data if not data.shape[0] == 0]
            n_samples = TF_data[0].shape[0]
            variables = {'yspec':np.transpose(TF_data),'freq':float(sampling_rate),
                         'dt2' :[0,float(len(order)-1),0],'npts':float((n_samples-1)*2),
                         'tfun':1}
            time_series_fname = file[:-4]+'_TF.mat'
            sio.savemat(time_series_fname,variables,appendmat = False)
        # Save Sonogram
        if 'sonogram' in var_names:
            sono_data = channel_set.channel_data(order[0],'sonogram')
            if not sono_data.shape[0] == 0:
                sono_phase = channel_set.channel_data(order[0],'sonogram_phase')
                sono_step = channel_set.channel_data(order[0],'sonogram_step')
                n_samples = sono_data[0].shape[0]
                variables = {'yson':to_dB(np.abs(sono_data)),'freq':float(sampling_rate),
                             'dt2' :[0,0,1],'npts':float((n_samples-1)*2),
                             'sonstep':float(sono_step),'yphase': sono_phase}
                time_series_fname = file[:-4]+'_sonogram.mat'
                sio.savemat(time_series_fname,variables,appendmat = False)
            
class DataExportWidget(QWidget):
    """
    A proof-of-concept widget to show that exporting data is possible.
    Currently, it saves all of the available variables. Alpha stage.
    
    Attributes
    ----------
    cs : ChannelSet
        Reference to the channelSet
    order : list
        The order of channels to be saved    
    """
    def __init__(self,parent):
        super().__init__(parent)
        self.cs = ChannelSet()
        self.order = list(range(len(self.cs)))
        self.init_UI()
        
    def init_UI(self):
        """
        Construct the UI
        """
        layout = QVBoxLayout(self)
        
        self.channel_listview = QListWidget(self)
        layout.addWidget(self.channel_listview)
        shift_btn_layout = QHBoxLayout()
        shift_up_btn = QPushButton('Shift Up',self)
        shift_up_btn.clicked.connect(lambda: self.shift_list('up'))
        shift_down_btn = QPushButton('Shift Down',self)
        shift_down_btn.clicked.connect(lambda: self.shift_list('down'))
        shift_btn_layout.addWidget(shift_up_btn )
        shift_btn_layout.addWidget(shift_down_btn )
        self.pickle_export_btn = QPushButton('Export as Pickle',self)
        self.pickle_export_btn.clicked.connect(self.pickle_file)
        self.back_comp_btn = QCheckBox('Backward Compatibility?',self)
        self.mat_export_btn = QPushButton('Export as MAT',self)
        self.mat_export_btn.clicked.connect(self.export_files)
        
        # TODO: Have a preview of what is being saved?
        layout.addWidget(QLabel('ChannelSet Saving Order',self))
        layout.addWidget(self.channel_listview)
        layout.addLayout(shift_btn_layout)
        layout.addWidget(self.pickle_export_btn)
        layout.addWidget(self.back_comp_btn)
        layout.addWidget(self.mat_export_btn)
        
    def set_channel_set(self, channel_set):
        """
        Set the ChannelSet reference
        """
        self.cs = channel_set
        self.order = list(range(len(self.cs)))
        self.update_list()
        self.channel_listview.setCurrentRow(0)
    
    def shift_list(self,mode):
        """
        Shift the channel saving order
        
        Parameter
        ---------
        mode : str
            either 'up' or 'down', indicating the shifting direction
        """
        ind = self.channel_listview.currentRow()
        if mode == 'up':
            if not ind == 0:
                self.order[ind-1],self.order[ind] = self.order[ind],self.order[ind-1]
                ind -= 1
        elif mode == 'down':
            if not ind == self.channel_listview.count()-1:
                self.order[ind],self.order[ind+1] = self.order[ind+1],self.order[ind]
                ind += 1
        self.update_list()
        self.channel_listview.setCurrentRow(ind)
    
    def update_list(self):
        """
        Update the list with the current saving order
        """
        self.channel_listview.clear()
        names = self.cs.channel_metadata(tuple(self.order),'name')
        full_names = [str(self.order[i])+' : '+names[i] for i in range(len(names))]
        self.channel_listview.addItems(full_names)
        
    def export_files(self):
        """
        Export the file to the url selected
        """
        url = QFileDialog.getSaveFileName(self, "Export Data", "",
                                           "MAT Files (*.mat)")[0]
        if url:
            if self.back_comp_btn.checkState() == Qt.Checked:
                export_to_mat(url,tuple(self.order), self.cs,back_comp = True)
            else:
                export_to_mat(url,tuple(self.order), self.cs)
      
    def pickle_file(self):
        '''
        This is probably a temporary solution to saving data.
        The pickle file is only intended for internal use only.
        Please make sure no one tampers with the files.
        Also, only open files that YOU have pickled.
        '''
        url = QFileDialog.getSaveFileName(self, "Export Data", "",
                                           "Pickle Files (*.pickle)")[0]
        if url:
            saved_cs = ChannelSet(len(self.order))
            for i,n in enumerate(self.order):
                saved_cs.channels[i] = self.cs.channels[n]
            with open(url,'wb') as f:
                pickle.dump(saved_cs,f)
    
if __name__ == '__main__':
    cs = ChannelSet(4)
    cs.add_channel_dataset((0,1,2,3),'time_series',np.random.rand(5,1))
    cs.set_channel_metadata(0,{'name':'Nope'})
    cs.set_channel_metadata(1,{'name':'Lol'})
    export_to_mat('btest.mat',(0,1,2,3),cs,True)




from cued_datalogger.api.pyqtgraph_extensions import InteractivePlotWidget
from cued_datalogger.api.toolbox import Toolbox
from cued_datalogger.api.numpy_extensions import to_dB
from cued_datalogger.api.channel import ChannelSet

from PyQt5.QtWidgets import QWidget, QGridLayout, QPushButton, QComboBox,QCheckBox
from PyQt5.QtCore import pyqtSignal,Qt

import scipy.fftpack
import numpy as np
from numpy.fft import rfft
import pyqtgraph as pg


class FrequencyDomainWidget(InteractivePlotWidget):
    """
    The FrequencyDomainWidget is the main display widget for everything in
    the frequency domain.
    """
    def __init__(self, parent=None):
        super().__init__(parent)

        self.plot_types = ['Linear Magnitude',
                           'Log Magnitude',
                           'Phase',
                           'Real Part',
                           'Imaginary Part',
                           'Nyquist']
        self.current_plot_type = self.plot_types[0]

        self.current_plot = "spectrum"

        self.coherence_plot = False

    def set_selected_channels(self, selected_channels):
        """Update which channels are plotted."""
        # If no channel list is given
        if not selected_channels:
            self.channels = []
        else:
            self.channels = selected_channels

        self.update_plot()

    def set_view_type(self,vtype):
        self.current_plot = vtype

    def set_plot_type(self, index):
        self.current_plot_type = self.plot_types[index]
        self.update_plot()

    def switch_cor_plot(self,state):
        if state == Qt.Unchecked:
            self.coherence_plot = False
        elif state == Qt.Checked:
            self.coherence_plot = True
        self.update_plot()

    def update_plot(self):
        self.clear()
        print("Plotting %s %s." % (self.current_plot_type,self.current_plot))
        for channel in self.channels:

            data = channel.data(self.current_plot)
            if data.shape[0] == 0:
                print('No %s to plot' % self.current_plot)
                continue

            # Plot
            if self.current_plot_type == 'Linear Magnitude':
                self.plot(channel.data("frequency"),
                          np.abs(data),
                          pen=pg.mkPen(channel.colour))

            elif self.current_plot_type == 'Log Magnitude':
                self.plot(channel.data("frequency"),
                          to_dB(np.abs(data)),
                          pen=pg.mkPen(channel.colour))

            elif self.current_plot_type == 'Phase':
                self.plot(channel.data("frequency"),
                          np.angle(data,deg=True),
                          pen=pg.mkPen(channel.colour))

            elif self.current_plot_type == 'Real Part':
                self.plot(channel.data("frequency"),
                          np.real(data),
                          pen=pg.mkPen(channel.colour))

            elif self.current_plot_type == 'Imaginary Part':
                self.plot(channel.data("frequency"),
                          np.imag(data),
                          pen=pg.mkPen(channel.colour))

            elif self.current_plot_type == 'Nyquist':
                self.plot(np.real(data),
                          np.imag(data),
                          pen=pg.mkPen(channel.colour))

            if self.current_plot == "TF" and self.coherence_plot:
                cor = channel.data("coherence")
                if cor.shape[0]:
                    self.plot(channel.data("frequency"),
                          cor,
                          pen=pg.mkPen('k'))

class FrequencyToolbox(Toolbox):
    """Toolbox containing the Frequency Domain controls."""
    sig_convert_to_circle_fit = pyqtSignal()
    sig_plot_type_changed = pyqtSignal(int)
    sig_view_type_changed = pyqtSignal(str)
    sig_coherence_plot = pyqtSignal(int)
    sig_converted_TF = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.parent = parent
        self.cs = None

        self.init_ui()

    def init_ui(self):
        # # Plot options tab
        self.plot_options_tab = QWidget(self)
        plot_options_tab_layout = QGridLayout()
        self.view_type_combobox = QComboBox(self)
        self.view_type_combobox.addItems(['Fourier Transform','Transfer Function'])
        self.view_type_combobox.currentIndexChanged[str].connect(self.sig_view_type_changed.emit)
        plot_options_tab_layout.addWidget(self.view_type_combobox, 0, 0)
        self.coherence_plot_tickbox = QCheckBox('Plot Coherence',self)
        self.coherence_plot_tickbox.stateChanged.connect(self.sig_coherence_plot.emit)
        plot_options_tab_layout.addWidget(self.coherence_plot_tickbox, 1, 0)
        self.plot_type_combobox = QComboBox(self.plot_options_tab)
        self.plot_type_combobox.addItems(['Linear Magnitude',
                                          'Log Magnitude',
                                          'Phase',
                                          'Real Part',
                                          'Imaginary Part',
                                          'Nyquist'])

        self.plot_type_combobox.setCurrentIndex(0)
        self.plot_type_combobox.currentIndexChanged.connect(self.sig_plot_type_changed.emit)
        plot_options_tab_layout.addWidget(self.plot_type_combobox, 2, 0)

        plot_options_tab_layout.setRowStretch(3, 1)
        self.plot_options_tab.setLayout(plot_options_tab_layout)

        self.addTab(self.plot_options_tab, "Plot Options")
        # # Conversion tab
        self.convert_tab = QWidget(self)
        convert_tab_layout = QGridLayout()

        self.view_tf_btn = QPushButton("Convert FFT to TF")
        #self.view_tf_btn.clicked.connect(self.sig_convert_to_TF.emit)
        self.view_tf_btn.clicked.connect(self.compute_tf)
        convert_tab_layout.addWidget(self.view_tf_btn, 0, 0)

        self.circle_fit_btn = QPushButton("Convert to Circle Fit")
        self.circle_fit_btn.clicked.connect(self.sig_convert_to_circle_fit.emit)
        convert_tab_layout.addWidget(self.circle_fit_btn, 1, 0)

        convert_tab_layout.setRowStretch(2, 1)
        self.convert_tab.setLayout(convert_tab_layout)

        self.addTab(self.convert_tab, "Conversion")

    def set_view_type(self,vtype):
        self.view_type_combobox.setCurrentText(vtype)

    def set_channel_set(self,cs):
        self.cs = cs

    def compute_tf(self):
        # If no TF exists, calculate one
        print("Calculating TF...")
        fdata = self.cs.channel_data(tuple(range(len(self.cs))),"spectrum")
        chans = list(range(len(self.cs)))
        in_chan = 0
        chans.remove(in_chan)
        input_chan_data = fdata[in_chan]
        autospec_in = compute_autospec(input_chan_data)

        for chan in chans:
            autospec_out = compute_autospec(fdata[chan])
            crossspec = compute_crossspec(input_chan_data,fdata[chan])
            tf,_ = compute_transfer_function(autospec_in,autospec_out,crossspec)
            if not self.cs.channels[chan].is_dataset('TF'):
                self.cs.add_channel_dataset(chan, 'TF', tf)
            else:
                self.cs.set_channel_data(chan, 'TF', tf)
        print("Done.")

        self.sig_converted_TF.emit()


def compute_autospec(fft_data):
    return(fft_data * np.conjugate(fft_data))

def compute_crossspec(in_fft_data,out_fft_data):
    return(np.conjugate(in_fft_data) * out_fft_data)

def compute_transfer_function(autospec_in,autospec_out,crossspec):

    transfer_func = (autospec_out/crossspec)
    coherence = ((crossspec * np.conjugate(crossspec))/(autospec_in*autospec_out))
    return(transfer_func,np.real(coherence))





from cued_datalogger.api.pyqtgraph_extensions import InteractivePlotWidget
from cued_datalogger.api.toolbox import Toolbox

from PyQt5.QtWidgets import QWidget, QGridLayout, QPushButton
from PyQt5.QtCore import pyqtSignal

import pyqtgraph as pg
import numpy as np
from numpy.fft import rfft

class TimeDomainWidget(InteractivePlotWidget):
    """
    The TimeDomainWidget is the main display widget for everything in
    the time domain.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.channels = []

    def set_selected_channels(self, selected_channels):
        """Update which channels are plotted"""
        # If no channel list is given
        if not selected_channels:
            self.channels = []
        else:
            self.channels = selected_channels
        self.update_plot()

    def update_plot(self):
        self.clear()
        for channel in self.channels:
            time_series = channel.data("time_series")
            if not time_series.shape[0] == 0:
                self.plot(channel.data("time"),
                          time_series,
                          pen=channel.colour)


class TimeToolbox(Toolbox):
    """Toolbox containing the Time Domain controls."""

    sig_convert_to_sonogram = pyqtSignal()
    sig_convert_to_fft = pyqtSignal()
    sig_converted_FFT = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.parent = parent
        self.cs = None

        self.init_ui()

    def init_ui(self):
        self.convert_tab = QWidget(self)
        convert_tab_layout = QGridLayout()

        self.fft_btn = QPushButton("Convert to FFT")
        self.fft_btn.clicked.connect(self.compute_fft)
        convert_tab_layout.addWidget(self.fft_btn, 0, 0)

        self.sonogram_btn = QPushButton("Convert to Sonogram")
        self.sonogram_btn.clicked.connect(self.sig_convert_to_sonogram.emit)
        convert_tab_layout.addWidget(self.sonogram_btn, 1, 0)

        convert_tab_layout.setRowStretch(2, 1)

        self.convert_tab.setLayout(convert_tab_layout)

        self.addTab(self.convert_tab, "Conversion")

    def set_channel_set(self,cs):
        self.cs = cs

    def compute_fft(self):
        # If no spectrum exists, calculate one
        print("Calculating spectrum...")
        for i in range(len(self.cs)):
            time_sig = self.cs.channel_data(i,"time_series")
            if not time_sig.shape[0] == 0:
                #window = np.hanning(len(time_sig))
                cycle=np.linspace(0,2*np.pi,time_sig.shape[0])
                window=1-np.cos(cycle);
                spectrum = rfft(time_sig * window)
                if not self.cs.channels[i].is_dataset("spectrum"):
                    self.cs.add_channel_dataset(i, "spectrum", spectrum)
                else:
                    self.cs.set_channel_data(i, "spectrum", spectrum)
            else:
                print('No Time Signal')

        print("Done.")
        self.sig_converted_FFT.emit()




"""
This module contains the class to record data from a soundcard.
It uses PyAudio to do so.
Please check the PyAudio Documentation for more information.

Typical example of using the module:
>>>import myRecorder as mR
>>>recorder = mR.Recorder()
Channels: 1
Rate: 44100
Chunk size: 1024
Number of chunks: 4
You are using pyAudio for recording
Device not found, reverting to default
Selected device: Line (3- U24XL with SPDIF I/O)
>>>recorder.stream_init()
stream already started
Input latency: 2.322e-02
Output latency: 0.000e+00
Read Available: -9977
Write Available: -9976
True
>>>recorder.record_init()
Recording function is ready! Use record_start() to start
True
>>>recorder.record_start()
stream already started
Recording Start!
True
>>>Recording Done! Please flush the data with flush_record_data().
data = recorder.flush_record_data()
>>>recorder.close()
"""

from cued_datalogger.acquisition.RecorderParent import RecorderParent
import sys,traceback

try:
    import pyaudio
except ImportError:
    # If pyaudio doesn't work, create mock version of it
    # If PyDAQmx doesn't work, create mock version of it
    if sys.version_info >= (3,):
        from unittest.mock import MagicMock
    else:
        from mock import Mock as MagicMock
    class MockModule(MagicMock):
        @classmethod
        def __getattr__(cls, name):
                return MagicMock()

    sys.modules['pyaudio'] = MockModule()
    import pyaudio

import numpy as np
import pprint as pp

class Recorder(RecorderParent):
    """
    Sets up the recording stream through a SoundCard

    Attributes
    ----------
    device_index: int
        Index of the device to be used for recording
    device_name: str
        Name of the device to be used for recording
    max_value: float
        Maximum value of recorded data
    """

#---------------- INITIALISATION METHODS -----------------------------------
    def __init__(self,channels = 1,rate = 44100, chunk_size = 1024,
                 num_chunk = 4,device_name = None):
        """
         Re-implemented from RecorderParent
        """

        super().__init__(channels = channels,rate = rate,
             chunk_size = chunk_size,num_chunk = num_chunk)

        print('You are using pyAudio for recording')
        self.p = None
        self.format = pyaudio.paInt16
        self.device_index = None;
        self.device_name = device_name

        self.open_recorder()
        self.set_device_by_name(str(device_name))

        self.trigger_init()

        self.max_value = 1;

    def open_recorder(self):
        """
         Re-implemented from RecorderParent. Prepare the PyAudio Object too.
        """
        super().open_recorder()
        if self.p == None:
            self.p = pyaudio.PyAudio()

#---------------- DESTRUCTOR METHODS -----------------------------------
    def close(self):
        """
         Re-implemented from RecorderParent, but terminate the PyAudio Object too.
        """
        super().close()
        #self.stream_close()
        if not self.p:
            self.p.terminate()
            self.p = None

#---------------- DEVICE SETTING METHODS -----------------------------------
    def set_device_by_name(self, name):
        """
        Set the recording audio device by name.
        Revert to default if no such device found

        Parameters
        ----------
        name: str
            Name of the device
        """
        dev_name,dev_index = self.available_devices()
        if not dev_name:
            print("Seems like you don't have any input devices")
            return

        try:
            self._set_device_by_index(dev_index[dev_name.index(name)])
        except ValueError:
            try:
                print('Device not found, reverting to default')
                default = self.p.get_default_input_device_info()
                self._set_device_by_index(default['index'])
            except IOError:
                print('No default device!')
                try:
                    self._set_device_by_index(dev_index[0])
                except IOError:
                    print('No Device can be set')
        except IOError:
            print('Device chosen cannot be found!')

    def available_devices(self):
        """
        Searches for any available input devices

        Returns
        ----------
        names: List
            Name of the devices
        index: List
            Index of the devices
        """
        names = [self.p.get_device_info_by_index(i)['name']
                  for i in range(self.p.get_device_count())
                  if self.p.get_device_info_by_index(i)['maxInputChannels']>0]

        index = [self.p.get_device_info_by_index(i)['index']
                for i in range(self.p.get_device_count())
                if self.p.get_device_info_by_index(i)['maxInputChannels']>0]

        return(names,index)

    def current_device_info(self):
        """
        Display the current selected device info
        """
        if self.device_index:
            pp.pprint(self.p.get_device_info_by_index(self.device_index))
        else:
            print('No index set')

    def _set_device_by_index(self,index):
        """
        Set the selected device by index

        Parameters
        ----------
        index: int
            Index of the device to be set
        """
        self.device_index = index;
        self.device_name = self.p.get_device_info_by_index(index)['name']
        print("Selected device: %s" % self.device_name)

#---------------- DATA METHODS -----------------------------------
    # Convert data obtained into a proper array
    def audiodata_to_array(self,data):
        """
        Re-implemented from RecorderParent
        """
        return np.frombuffer(data, dtype = np.int16).reshape((self.chunk_size,self.channels))/ 2**15

#---------------- STREAMING METHODS -----------------------------------
    def stream_audio_callback(self,in_data, frame_count, time_info, status):
        """
        Callback function for audio streaming.
        First, it writes data to the circular buffer,
        then record data if it is recording,
        finally check for any trigger.

        Inputs and Outputs are part of the callback format.
        More info can be found in PyAudio documentation
        """
        data_array = self.audiodata_to_array(in_data)
        self.write_buffer(data_array)
        #self.rEmitter.newdata.emit()

        if self.recording:
            self.record_data(data_array)

         # Trigger check
        if self.trigger:
            self._trigger_check_threshold(data_array)

        return(in_data,pyaudio.paContinue)

    # TODO: Check for valid device, channels and all that before initialisation
    def stream_init(self, playback = False):
        """
        Re-implemented from RecorderParent.
        """
        if (not self.device_index == None) and (self.audio_stream == None) :
            try:
                self.audio_stream = self.p.open(channels = self.channels,
                                 rate = self.rate,
                                 format = self.format,
                                 input = True,
                                 output = playback,
                                 frames_per_buffer = self.chunk_size,
                                 input_device_index = self.device_index,
                                 stream_callback = self.stream_audio_callback)

                self.stream_start()
                print('Input latency: %.3e' % self.audio_stream.get_input_latency())
                print('Output latency: %.3e' % self.audio_stream.get_output_latency())
                print('Read Available: %i' % self.audio_stream.get_read_available())
                print('Write Available: %i' % self.audio_stream.get_write_available())
                return True

            except:
                t,v,tb = sys.exc_info()
                print(t)
                print(v)
                print(traceback.format_tb(tb))
                self.audio_stream = None
                return False
        else:
            return False

    # Start the streaming
    def stream_start(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream:
            if self.audio_stream.is_stopped():
                self.audio_stream.start_stream()
            else:
                print('stream already started')
        else:
            print('No audio stream is set up')

    # Stop the streaming
    def stream_stop(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream:
            if not self.audio_stream.is_stopped():
                self.audio_stream.stop_stream()
            else:
                print('stream already stopped')
        else:
            print('No audio stream is set up')

    # Close the stream, probably needed if any parameter of the stream is changed
    def stream_close(self):
        """
        Re-implemented from RecorderParent.
        """
        if self.audio_stream and self.audio_stream.is_active():
            self.stream_stop()
            self.audio_stream.close()
            self.audio_stream = None




from setuptools import setup
import sys
import subprocess
from os.path import isfile
import urllib.request
import traceback
import pip

def version():
    """Get version number"""
    with open('cued_datalogger/VERSION') as f:
        return f.read()

def readme():
    """Get text from the README.rst"""
    with open('README.rst') as f:
        return f.read()

error_messages = {}

# Check for Anaconda
print("Checking for Anaconda installation...")
if "Anaconda" in sys.version or "Continuum" in sys.version:
    print("\t Anaconda installation found.\n")
    use_anaconda = True
else:
    print("\t Anaconda not found. [OK]\n")
    use_anaconda = False


# Check OS
if sys.platform.startswith("win"):
    operating_system = "windows"
elif sys.platform.startswith("linux"):
    operating_system = "linux"
elif sys.platform == "darwin":
    operating_system = "os x"
else:
    operating_system = "unable to detect operating system"

print("Operating system: {}.\n".format(operating_system))

if operating_system == "unable to detect operating system":
    error_messages["operating_system"] = \
        ("[WARNING] The DataLogger has not been configured for this operating"
          " system. \n The installation may not work correctly.\n")
    print(error_messages["operating_system"])


# Do some OS / Anaconda specific config
if operating_system != "windows":
    # Find portaudio:
    print("Searching for portaudio...")
    portaudio = subprocess.check_output(['whereis', 'portaudio']).decode("utf-8").split()
    if len(portaudio) > 1:
        print("\t Portaudio found at {}".format(portaudio[1]))
    else:
        error_messages["portaudio"] = \
            ("\t Portaudio not installed - please install it manually using the "
              "appropriate package manager for your system: \n"
              "\t \t OS X: brew install portaudio \n"
              "\t \t Ubuntu/Debian: sudo apt install libportaudio2 portaudio19-dev \n"
              "\t \t CentOS: yum install portaudio portaudio-devel")
        print(error_messages["portaudio"])

    if use_anaconda:
        # Find conda
        conda_executable = subprocess.check_output(['which', 'conda']).decode("utf-8").split()[0]
        print("Conda executable found at {}".format(conda_executable))

if use_anaconda and operating_system == "windows":
    # Find conda
    conda_executable = subprocess.check_output(['where', 'conda']).decode("utf-8").split()
    if isinstance(conda_executable, list):
        for item in conda_executable:
            if item.endswith(".bat") or item.endswith(".exe"):
                conda_executable = item
                break
    print("Conda executable found at {}".format(conda_executable))

    # Check for python3.dll
    print("Searching for python3.dll...")
    print("\t(In Anaconda on Windows, this dll is required for the DataLogger"
          " UI)")

    python_dll_found = False

    # Find where the python executables are (returns a list)
    python_path = subprocess.check_output(['where', 'python']).decode("utf-8")
    python_path = python_path.split()

    anaconda_paths = []

    # Process these to get just the anaconda paths
    for location in python_path:
        # Get the first part of the path
        # eg. \home\Anaconda\python3\ becomes \home\Anaconda
        anaconda_path = location.split('python')[0]
        # Swap the \ for /
        anaconda_path = anaconda_path.replace("\\", '/')
        anaconda_paths.append(anaconda_path)

    # Look in all the conda paths for the .dll
    for location in anaconda_paths:
        if isfile(location + "Library/bin/python3.dll"):
            print("python3.dll found in " + location
                  + "Library/bin/python3.dll")
            python_dll_found = True
            break

    if python_dll_found:
        print("\t python3.dll file found, so DataLogger will work "
              "with this Anaconda installation.\n")
    else:
        # TODO: Currently this script installs the python3.dll automatically
        # with no user input prompting. It might be nice to have it prompt.
        print("No python3.dll found.")
        dll_url = ("https://github.com/torebutlin/cued_datalogger/tree/"
                   "master/lib/python3.dll")

        print("Downloading from {}...".format(dll_url))
        for location in anaconda_paths:
            # Download the dll
            try:
                urllib.request.urlretrieve(dll_url,
                                           location + 'Library/bin/python3.dll')
            except:
                print("Error downloading python3.dll.")
                traceback.print_exc()
                sys.exit(1)
            print("python3.dll downloaded to {}.\n".format(location))


print("Configuring dependency list...")
if use_anaconda:
    conda_dependency_list = ['numpy', 'scipy']
    dependency_list = ['matplotlib',
                       'pyaudio',
                       'pydaqmx',
                       'pyqt5',
                       'pyqtgraph']
else:
    dependency_list = ['numpy',
                       'scipy',
                       'matplotlib',
                       'pyaudio',
                       'pydaqmx',
                       'pyqt5',
                       'pyqtgraph']
print("\tDependencies: ")
for package in dependency_list:
    print("\t \t " + package)
if use_anaconda:
    for item in conda_dependency_list:
        print("\t \t " + item + (" (conda version)"))


print("\nInstalling dependencies...")

if use_anaconda:
    # Install anaconda dependencies
    print("Installing conda dependencies... ")
    for dependency in conda_dependency_list:
        print("Installing " + dependency + " using conda:")
        install_command = [conda_executable, 'install', '-y', dependency]
        conda_install_process = subprocess.Popen(install_command, stdout=subprocess.PIPE, stdin=subprocess.PIPE, stderr=subprocess.STDOUT)
        #while conda_install_process.stdout:
        for line in conda_install_process.stdout:
            print(line.decode("utf-8")[:-1])

# PyQt5 does not have a source distribution currently (09/2017) so it cannot
# be installed using setuptools. The wheel must be installed manually.
# see https://stackoverflow.com/questions/4628519/is-it-possible-to-require-pyqt-from-setuptools-setup-py/45598092#45598092

# There were also problems with pyaudio depending on some c libraries.
# As a result, it was decided to install all the packages this way rather than
# using the setuptools install_requires parameter.
for package in dependency_list:
    pip.main(['install', '-v', package])


print("\nContinuing to setup...\n")

setup(name='cued_datalogger',
      version=version(),
      description='The CUED DataLogger for acquiring and analysing data',
      long_description=readme(),
      url='https://github.com/torebutlin/cued_datalogger',
      author='Theo Brown, En Yi Tee',
      author_email='tab53@cam.ac.uk, eyt21@cam.ac.uk',
      license='BSD 3-Clause License',
      packages=['cued_datalogger',
                'cued_datalogger/acquisition',
                'cued_datalogger/analysis',
                'cued_datalogger/api'],
      install_requires=None,#dependency_list,
      entry_points={
        'console_scripts': ['cued_datalogger_dbg ='
                            ' cued_datalogger.__main__:run'],
        'gui_scripts': ['cued_datalogger = '
                        'cued_datalogger.__main__:run']},
      zip_safe=True,
      include_package_data=True)

if error_messages:
    print("Installation failed with the following errors:")
    for error, message in error_messages.items():
        print("Error: " + message)



import numpy as np
from numpy.fft import rfft,fft

from PyQt5.QtWidgets import (QWidget, QApplication, QTabWidget, QComboBox,
                             QHBoxLayout, QMainWindow, QPushButton,
                             QVBoxLayout, QAction, QMenu)

import sys

from cued_datalogger.analysis.circle_fit import CircleFitWidget, CircleFitToolbox
from cued_datalogger.analysis.frequency_domain import FrequencyDomainWidget, FrequencyToolbox
from cued_datalogger.analysis.sonogram import SonogramDisplayWidget, SonogramToolbox
from cued_datalogger.analysis.time_domain import TimeDomainWidget, TimeToolbox

from cued_datalogger.api.addons import AddonManager
from cued_datalogger.api.channel import ChannelSet, ChannelSelectWidget, ChannelMetadataWidget
from cued_datalogger.api.file_import import DataImportWidget
from cued_datalogger.api.file_export import DataExportWidget
from cued_datalogger.api.toolbox import Toolbox, MasterToolbox

from cued_datalogger.acquisition.acquisition_window import AcquisitionWindow
from cued_datalogger.acquisition.RecordingUIs import DevConfigUI

class AnalysisDisplayTabWidget(QTabWidget):
    """This is the central widget for the AnalysisWindow, where graphs, data,
    and results are displayed. In general the user does not interact explicitly
    with this widget, instead accessing the widgets it contains.

    Attributes
    ----------
    circle_widget : :class:`~cued_datalogger.analysis.circle_fit.CircleFitWidget`
    freqdomain_widget : :class:`~cued_datalogger.analysis.frequency_domain.FrequencyDomainWidget`
    sonogram_widget : :class:`~cued_datalogger.analysis.sonogram.SonogramDisplayWidget`
    timedomain_widget : :class:`~cued_datalogger.analysis.time_domain.TimeDomainWidget`
    """
    def __init__(self, *arg, **kwarg):
        super().__init__(*arg, **kwarg)

        self.init_ui()

    def init_ui(self):
        self.setMovable(False)
        self.setTabsClosable(False)

        self.timedomain_widget = TimeDomainWidget(self)

        self.freqdomain_widget = FrequencyDomainWidget(self)
        self.sonogram_widget = SonogramDisplayWidget(self)
        self.circle_widget = CircleFitWidget(self)

        # Create the tabs
        self.addTab(self.timedomain_widget, "Time Domain")
        self.addTab(self.freqdomain_widget, "Frequency Domain")
        self.addTab(self.sonogram_widget, "Sonogram")
        self.addTab(self.circle_widget, "Circle Fit")


class ProjectMenu(QMenu):
    '''
    A simple drop-down menu for demonstration purposes.
    Currently does not do anything.
    '''
    def __init__(self,parent):
        super().__init__('Project',parent)
        self.parent = parent
        self.initMenu()

    def initMenu(self):
        newAct = QAction('&New', self)
        newAct.setShortcut('Ctrl+N')

        setAct = QAction('&Settings', self)

        exitAct = QAction('&Exit', self)
        exitAct.setShortcut('Ctrl+Q')
        exitAct.setStatusTip('Exit application')

        self.addActions([newAct,setAct,exitAct])


class AnalysisWindow(QMainWindow):
    """
    The main window for analysing and processing data.

    Attributes
    ----------
    cs : :class:`~cued_datalogger.api.channel.ChannelSet`
      The ChannelSet containing all the data. Data is accessed through this
      ChannelSet.

    menubar : :class:`PyQt5.QtWidgets.QMenuBar`

    toolbox : :class:`~cued_datalogger.api.toolbox.MasterToolbox`
      The widget containing local tools and operations. Contains four
      toolboxes: :attr:`time_toolbox`, :attr:`frequency_toolbox`,
      :attr:`sonogram_toolbox`, :attr:`circle_fit_toolbox`.

    time_toolbox : :class:`~cued_datalogger.analysis.time_domain.TimeToolbox`
    frequency_toolbox : :class:`~cued_datalogger.analysis.frequency_domain.FrequencyToolbox`
    sonogram_toolbox : :class:`~cued_datalogger.analysis.sonogram.SonogramToolbox`
    circle_fit_toolbox : :class:`~cued_datalogger.analysis.circle_fit.CircleFitToolbox`

    display_tabwidget : :class:`~cued_datalogger.analysis_window.AnalysisDisplayTabWidget`
        The central widget for display.

    global_master_toolbox : :class:`~cued_datalogger.api.toolbox.MasterToolbox`
      The master toolbox containing the :attr:`global_toolbox`.

    global_toolbox : :class:`~cued_datalogger.api.toolbox.Toolbox`
      The widget containing global tools and operations. Has five tabs,
      containing: <acquisition window launcher>, :attr:`channel_select_widget`,
      :attr:`channel_metadata_widget`, :attr:`addon_widget`,
      :attr:`import_widget`.
    acquisition_window : :class:`~cued_datalogger.acquisition_window.AcquisitionWindow`
    channel_select_widget : :class:`~cued_datalogger.api.channel.ChannelSelectWidget`
    channel_metadata_widget : :class:`~cued_datalogger.api.channel.ChannelMetadataWidget`
    addon_widget : :class:`~cued_datalogger.api.addons.AddonManager`
    import_widget : :class:`~cued_datalogger.api.file_import.DataImportWidget`
    """
    def __init__(self):
        super().__init__()

        self.setGeometry(500,300,800,500)
        self.setWindowTitle('AnalysisWindow')

        #self.new_cs = ChannelSet()
        self.create_test_channelset()
        self.init_ui()

        self.setFocus()
        self.show()

    def init_toolbox(self):
        """Create the master toolbox"""
        self.toolbox = MasterToolbox(self)

        # # Time toolbox
        self.time_toolbox = TimeToolbox(self.toolbox)
        self.time_toolbox.sig_converted_FFT.connect(self.plot_fft)
        self.time_toolbox.sig_convert_to_sonogram.connect(self.plot_sonogram)

        # # Frequency toolbox
        self.frequency_toolbox = FrequencyToolbox(self.toolbox)
        self.frequency_toolbox.set_channel_set(self.cs)
        self.frequency_toolbox.sig_plot_type_changed.connect(self.display_tabwidget.freqdomain_widget.set_plot_type)
        self.frequency_toolbox.sig_view_type_changed.connect(self.set_freq_plot_type)
        self.frequency_toolbox.sig_converted_TF.connect(self.plot_tf)
        self.frequency_toolbox.sig_coherence_plot.connect(self.display_tabwidget.freqdomain_widget.switch_cor_plot)
        self.frequency_toolbox.sig_convert_to_circle_fit.connect(self.circle_fitting)

        # # Sonogram toolbox
        self.sonogram_toolbox = SonogramToolbox(self.toolbox)
        self.sonogram_toolbox.sig_contour_spacing_changed.connect(self.display_tabwidget.sonogram_widget.update_contour_spacing)
        self.sonogram_toolbox.sig_num_contours_changed.connect(self.display_tabwidget.sonogram_widget.update_num_contours)
        self.sonogram_toolbox.sig_window_overlap_fraction_changed.connect(self.display_tabwidget.sonogram_widget.update_window_overlap_fraction)
        self.sonogram_toolbox.sig_window_width_changed.connect(self.display_tabwidget.sonogram_widget.update_window_width)

        # # Circle Fit toolbox
        self.circle_fit_toolbox = CircleFitToolbox(self.toolbox)
        self.circle_fit_toolbox.sig_show_transfer_fn.connect(self.display_tabwidget.circle_widget.show_transfer_fn)
        self.circle_fit_toolbox.sig_construct_transfer_fn.connect(self.display_tabwidget.circle_widget.construct_transfer_fn)
        #self.circle_fit_toolbox.sig_autofit_parameter_change.connect(self.display_tabwidget.circle_widget.update_autofit_parameters)

        self.toolbox.add_toolbox(self.time_toolbox)
        self.toolbox.add_toolbox(self.frequency_toolbox)
        self.toolbox.add_toolbox(self.sonogram_toolbox)
        self.toolbox.add_toolbox(self.circle_fit_toolbox)
        self.toolbox.set_toolbox(0)

    def init_global_master_toolbox(self):
        self.global_master_toolbox = MasterToolbox()

        self.global_toolbox = Toolbox('right', self.global_master_toolbox)

        # # Acquisition Window
        self.acquisition_window = None
        self.dev_configUI = DevConfigUI()
        self.dev_configUI.config_button.setText('Open AcquisitionWindow')
        self.dev_configUI.config_button.clicked.connect(self.open_acquisition_window)
        self.global_toolbox.addTab(self.dev_configUI,'AcquisitionWindow')

        # # Channel Selection
        self.channel_select_widget = ChannelSelectWidget(self.global_toolbox)
        self.channel_select_widget.sig_channel_selection_changed.connect(self.display_tabwidget.timedomain_widget.set_selected_channels)
        self.channel_select_widget.sig_channel_selection_changed.connect(self.display_tabwidget.freqdomain_widget.set_selected_channels)
        self.channel_select_widget.sig_channel_selection_changed.connect(self.display_tabwidget.sonogram_widget.set_selected_channels)
        self.channel_select_widget.sig_channel_selection_changed.connect(self.sonogram_toolbox.set_selected_channels)
        self.channel_select_widget.sig_channel_selection_changed.connect(self.display_tabwidget.circle_widget.set_selected_channels)

        self.global_toolbox.addTab(self.channel_select_widget, 'Channel Selection')

        # # Channel Metadata
        self.channel_metadata_widget = ChannelMetadataWidget(self.global_toolbox)
        self.global_toolbox.addTab(self.channel_metadata_widget, 'Channel Metadata')

        self.channel_metadata_widget.metadataChange.connect(self.update_cs)

        # # Addon Manager
        self.addon_widget = AddonManager(self)
        self.global_toolbox.addTab(self.addon_widget, 'Addon Manager')

        # # Import
        self.import_widget = DataImportWidget(self)
        self.import_widget.add_data_btn.clicked.connect(lambda: self.add_import_data('Extend'))
        self.import_widget.rep_data_btn.clicked.connect(lambda: self.add_import_data('Replace'))
        self.global_toolbox.addTab(self.import_widget, 'Import Files')

        # # Export
        self.export_widget = DataExportWidget(self)
        self.global_toolbox.addTab(self.export_widget, 'Export Files')

        self.global_master_toolbox.add_toolbox(self.global_toolbox)

    def update_cs(self, cs):
        self.cs = cs
        self.channel_select_widget.cs = self.cs
        self.channel_select_widget.set_channel_name()

    def init_ui(self):
        # Add the drop-down menu
        self.menubar = self.menuBar()
        self.menubar.addMenu(ProjectMenu(self))

        # # Create the main widget
        self.main_widget = QWidget(self)
        self.setCentralWidget(self.main_widget)
        self.main_layout = QHBoxLayout()
        self.main_widget.setLayout(self.main_layout)

        # Create the analysis tools tab widget
        self.display_tabwidget = AnalysisDisplayTabWidget(self)

        # Create the toolbox
        self.init_toolbox()
        self.display_tabwidget.currentChanged.connect(self.toolbox.set_toolbox)

        # Create the global toolbox
        self.init_global_master_toolbox()

        self.tfplots = []
        self.configure_channelset()
        self.plot_time_series()
        #self.plot_fft()
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.timedomain_widget)

        # Add the widgets
        self.main_layout.addWidget(self.toolbox)
        self.main_layout.addWidget(self.display_tabwidget)
        self.main_layout.addWidget(self.global_master_toolbox)
        # Set the stretch factors
        self.main_layout.setStretchFactor(self.toolbox, 0)
        self.main_layout.setStretchFactor(self.display_tabwidget, 1)
        self.main_layout.setStretchFactor(self.global_master_toolbox, 0.5)

    def create_test_channelset(self):
        self.cs = ChannelSet(5)
        t = np.arange(0,0.5,1/5000)
        for i, channel in enumerate(self.cs.channels):
            self.cs.set_channel_metadata(i,{'sample_rate': 5000})
            self.cs.add_channel_dataset(i, 'time_series', np.sin(t*2*np.pi*100*(i+1)))
            self.cs.add_channel_dataset(i,'spectrum', [])
        self.cs.add_channel_dataset(i,'time_series', np.sin(t*2*np.pi*100*(i+1))*np.exp(-t/t[-1]) )
        self.cs.add_channel_dataset(i,'spectrum', [])

    def open_acquisition_window(self):
        if not self.acquisition_window:
            recType, configs = self.dev_configUI.read_device_config()
            if not any([not c for c in configs]):
                self.acquisition_window = AcquisitionWindow(self, recType, configs)
            else:
                self.acquisition_window = AcquisitionWindow(self)
            self.acquisition_window.done.connect(self.done_acquisition_window)
            self.acquisition_window.dataSaved.connect(self.receive_data)
            self.acquisition_window.show()

    def receive_data(self, tab_number=0):
        self.configure_channelset()
        self.plot_time_series()
        self.plot_fft()

        if self.sender() == self.acquisition_window:
            self.display_tabwidget.setCurrentIndex(tab_number)
            if tab_number == 1:
                self.set_freq_plot_type('Transfer Function')

    def done_acquisition_window(self):
        self.acquisition_window.done.disconnect()
        self.acquisition_window = None

    def configure_channelset(self):
        self.channel_select_widget.set_channel_set(self.cs)
        self.channel_metadata_widget.set_channel_set(self.cs)
        self.time_toolbox.set_channel_set(self.cs)
        self.frequency_toolbox.set_channel_set(self.cs)
        self.export_widget.set_channel_set(self.cs)

    def plot_time_series(self):
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.timedomain_widget)
        self.display_tabwidget.timedomain_widget.set_selected_channels(self.cs.channels)

    def set_freq_plot_type(self, plot_type):
        if plot_type == 'Fourier Transform':
            self.plot_fft()
        elif plot_type == 'Transfer Function':
            self.plot_tf()

    def plot_fft(self):
        # Switch to frequency domain tab
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.freqdomain_widget)
        self.display_tabwidget.freqdomain_widget.set_view_type("spectrum")
        self.display_tabwidget.freqdomain_widget.set_selected_channels(self.cs.channels)
        self.frequency_toolbox.set_view_type('Fourier Transform')

    def plot_tf(self):
        #TODO: calculate TF function if none is found
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.freqdomain_widget)
        self.display_tabwidget.freqdomain_widget.set_view_type("TF")
        self.display_tabwidget.freqdomain_widget.set_selected_channels(self.cs.channels)
        self.frequency_toolbox.set_view_type('Transfer Function')

    def plot_sonogram(self):
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.sonogram_widget)
        self.display_tabwidget.sonogram_widget.set_selected_channels(self.cs.channels)

    def circle_fitting(self):
        self.display_tabwidget.setCurrentWidget(self.display_tabwidget.circle_widget)
        self.display_tabwidget.circle_widget.set_selected_channels(self.cs.channels)

    def add_import_data(self,mode):
        if mode == 'Extend':
            self.cs.channels.extend(self.import_widget.new_cs.channels)
        elif mode == 'Replace':
            self.cs = self.import_widget.new_cs

        self.receive_data()
        self.import_widget.clear()


if __name__ == '__main__':
    app = 0
    app = QApplication(sys.argv)

    w = AnalysisWindow()
    w.show()

    sys.exit(app.exec_())
